# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

import clingo.ast
from ..utils import utils
from ..program_parser import transitive_closure
from ..program_parser import visitor


#
# DEFINES
#

# predicates
HOLDS      = utils.HOLDS
HOLDSP     = utils.HOLDSP
PREFERENCE = utils.PREFERENCE
OPTIMIZE   = utils.OPTIMIZE

# ems
M1    = visitor.M1
M2    = visitor.M2
M1_M2 = visitor.M1_M2
ZERO  = visitor.ZERO

# errors
ERROR_PROJECT    = utils.ERROR_PROJECT
ERROR_MINIMIZE   = utils.ERROR_MINIMIZE
ERROR_DISJOINT   = utils.ERROR_DISJOINT
ERROR_CSPLITERAL = utils.ERROR_CSPLITERAL
ERROR_HOLDSP     = utils.ERROR_HOLDSP


class BasicTermTransformer(visitor.TermTransformer):

    def __init__(self, underscores):
        visitor.TermTransformer.__init__(self)
        self.__underscores = underscores

    def set_predicates_info(self):
        self.predicates_info = dict()
        PredicateInfo = visitor.PredicateInfo
        self.predicates_info[(HOLDS, 1)]  = PredicateInfo(None, 0, ZERO, 1)
        info = PredicateInfo(None, 0, None, 0)
        for i in [(OPTIMIZE, 1), (PREFERENCE, 2), (PREFERENCE, 5)]:
            self.predicates_info[i] = info
        self.default = PredicateInfo(None, self.__underscores, None, 0)


class BasicProgramVisitor(visitor.Visitor):

    def __init__(self, builder, type, underscores, constants):
        visitor.Visitor.__init__(self)
        self.type = type
        self.helper = visitor.Helper()
        self.__builder = builder
        self.__underscores = underscores
        self.__constants = constants
        self.__in_program = False
        self.__term_transformer = BasicTermTransformer(underscores)
        self.__term_transformer.set_predicates_info()

    def __add(self, statement):
        if not self.__in_program:
            prg = clingo.ast.Program(statement.location, self.type, [])
            self.__builder.add(prg)
            self.__in_program = True
        self.__builder.add(statement)

    #
    # Statements
    #

    def visit_Rule(self, rule):
        self.visit_children(rule)
        self.__add(rule)

    def visit_Definition(self, d):
        if d.name not in self.__constants:
            self.__add(d)

    def visit_ShowSignature(self, sig):
        if sig.name != "" or sig.arity != 0:
            self.__term_transformer.transform_signature(sig)
        self.__add(sig)

    def visit_Defined(self, defined):
        if defined.name != "" or defined.arity != 0:
            self.__term_transformer.transform_signature(defined)
        self.__add(defined)

    def visit_ShowTerm(self, show):
        self.visit_children(show)
        show.term = self.__term_transformer.reify_term(show.term,
                                                       self.helper.show)
        self.__add(show)

    def visit_Minimize(self, min):
        self.visit_children(min)
        self.__add(min)

    def visit_Script(self, script):
        self.__add(script)

    def visit_Program(self, prg):
        pass # ignore

    def visit_External(self, ext):
        pass # ignore

    def visit_Edge(self, edge):
        self.visit_children(edge)
        tt = self.__term_transformer
        edge.u = tt.reify_term(edge.u, self.helper.edge + "_" + self.type)
        edge.v = tt.reify_term(edge.v, self.helper.edge + "_" + self.type)
        self.__add(edge)

    def visit_Heuristic(self, heur):
        self.visit_children(heur)
        self.__add(heur)

    def visit_ProjectAtom(self,atom):
        string = ERROR_PROJECT.format(self.type, str(atom))
        self.helper.raise_exception(string)

    def visit_ProjectSignature(self, sig):
        string = ERROR_PROJECT.format(self.type, str(sig))
        self.helper.raise_exception(string)


    #
    # Elements
    #

    def visit_SymbolicAtom(self, atom):
        if str(atom.term.type) == "Function":
            if atom.term.name == HOLDSP and len(atom.term.arguments)==1:
                string = ERROR_HOLDSP.format(self.type, str(atom))
                self.helper.raise_exception(string)
            self.__term_transformer.transform_function(atom.term)

    # csp literals are not accepted
    def visit_CSPLiteral(self,csp):
        string = ERROR_CSPLITERAL.format(self.type, str(csp))
        self.helper.raise_exception(string)

    # disjoint atoms are not accepted
    def visit_Disjoint(self,d):
        string = ERROR_DISJOINT.format(self.type, str(d))
        self.helper.raise_exception(string)


class HeuristicProgramVisitor(BasicProgramVisitor):

    def __init__(self, builder, type, underscores, consts):
        BasicProgramVisitor.__init__(self, builder, type, underscores, consts)

    def visit_Minimize(self, min):
        string = ERROR_MINIMIZE.format(self.type, str(min))
        self.helper.raise_exception(string)




meta_program = """
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This file is part of clingo.                                            %
% Copyright (C) 2015  Martin Gebser                                       %
% Copyright (C) 2015  Roland Kaminski                                     %
% Copyright (C) 2015  Torsten Schaub                                      %
%                                                                         %
% This program is free software: you can redistribute it and/or modify    %
% it under the terms of the GNU General Public License as published by    %
% the Free Software Foundation, either version 3 of the License, or       %
% (at your option) any later version.                                     %
%                                                                         %
% This program is distributed in the hope that it will be useful,         %
% but WITHOUT ANY WARRANTY; without even the implied warranty of          %
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           %
% GNU General Public License for more details.                            %
%                                                                         %
% You should have received a copy of the GNU General Public License       %
% along with this program.  If not, see <http://www.gnu.org/licenses/>.   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

##conjunction(B) :- ##literal_tuple(B),
        ##hold(L) : ##literal_tuple(B, L), L > 0;
    not ##hold(L) : ##literal_tuple(B,-L), L > 0.

##body(normal(B)) :- ##rule(_,normal(B)), ##conjunction(B).
##body(sum(B,G))  :- ##rule(_,sum(B,G)),
    #sum { W,L :     ##hold(L), ##weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : not ##hold(L), ##weighted_literal_tuple(B,-L,W), L > 0 } >= G.

  ##hold(A) : ##atom_tuple(H,A)   :- ##rule(disjunction(H),B), ##body(B).
{ ##hold(A) : ##atom_tuple(H,A) } :- ##rule(     choice(H),B), ##body(B).

% commented by Javier for MetaspPython
%*
##optimize(J,W,Q) :- ##output(_optimize(J,W,Q),B), ##conjunction(B).
 :- ##output(_query,B), not ##conjunction(B).

##hide(_criteria(J,W,Q)) :- ##output(_criteria(J,W,Q),_).
##hide(_query)           :- ##output(_query,_).
##hide(_optimize(J,W,Q)) :- ##output(_optimize(J,W,Q),_).
#show.
#show T : ##output(T,B), ##conjunction(B), not ##hide(T).
*%

% added by Javier for MetaspPython
#show.
#show T : ##output(T,L), ##hold(L).


"""

metaD_program = """

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This file is part of clingo.                                            %
%                                                                         %
% Authors: Martin Gebser, Roland Kaminski, Torsten Schaub                 %
%                                                                         %
% This program is free software: you can redistribute it and/or modify    %
% it under the terms of the GNU General Public License as published by    %
% the Free Software Foundation, either version 3 of the License, or       %
% (at your option) any later version.                                     %
%                                                                         %
% This program is distributed in the hope that it will be useful,         %
% but WITHOUT ANY WARRANTY; without even the implied warranty of          %
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           %
% GNU General Public License for more details.                            %
%                                                                         %
% You should have received a copy of the GNU General Public License       %
% along with this program.  If not, see <http://www.gnu.org/licenses/>.   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% NOTE: assumes that a rule has no more than one head

##sum(B,G,T) :- ##rule(_,sum(B,G)), T = #sum { W,L : ##weighted_literal_tuple(B,L,W) }.

% extract supports of atoms and facts

##supp(A,B) :- ##rule(     choice(H),B), ##atom_tuple(H,A).
##supp(A,B) :- ##rule(disjunction(H),B), ##atom_tuple(H,A).

##supp(A) :- ##supp(A,_).

##atom(|L|) :- ##weighted_literal_tuple(_,L,_).
##atom(|L|) :- ##literal_tuple(_,L).
##atom( A ) :- ##atom_tuple(_,A).

##fact(A) :- ##rule(disjunction(H),normal(B)), ##atom_tuple(H,A), not ##literal_tuple(B,_).

% generate interpretation

##true(atom(A))                :- ##fact(A).
##true(atom(A)); ##fail(atom(A)) :- ##supp(A), not ##fact(A).
                 ##fail(atom(A)) :- ##atom(A), not ##supp(A).

##true(normal(B)) :- ##literal_tuple(B),
    ##true(atom(L)) : ##literal_tuple(B, L), L > 0;
    ##fail(atom(L)) : ##literal_tuple(B,-L), L > 0.
##fail(normal(B)) :- ##literal_tuple(B, L), ##fail(atom(L)), L > 0.
##fail(normal(B)) :- ##literal_tuple(B,-L), ##true(atom(L)), L > 0.

##true(sum(B,G)) :- ##sum(B,G,T),
    #sum { W,L : ##true(atom(L)), ##weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : ##fail(atom(L)), ##weighted_literal_tuple(B,-L,W), L > 0 } >= G.
##fail(sum(B,G)) :- ##sum(B,G,T),
    #sum { W,L : ##fail(atom(L)), ##weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : ##true(atom(L)), ##weighted_literal_tuple(B,-L,W), L > 0 } >= T-G+1.

% verify supported model properties

##bot :- ##rule(disjunction(H),B), ##true(B), ##fail(atom(A)) : ##atom_tuple(H,A).
##bot :- ##true(atom(A)), ##fail(B) : ##supp(A,B).

% verify acyclic derivability

##internal(C,normal(B)) :- ##scc(C,A), ##supp(A,normal(B)), ##scc(C,A'), ##literal_tuple(B,A').
##internal(C,sum(B,G))  :- ##scc(C,A), ##supp(A,sum(B,G)),  ##scc(C,A'), ##weighted_literal_tuple(B,A',W).

##external(C,normal(B)) :- ##scc(C,A), ##supp(A,normal(B)), not ##internal(C,normal(B)).
##external(C,sum(B,G))  :- ##scc(C,A), ##supp(A,sum(B,G)),  not ##internal(C,sum(B,G)).

##steps(C,Z-1) :- ##scc(C,_), Z = { ##scc(C,A) : not ##fact(A) }.

##wait(C,atom(A),0)   :- ##scc(C,A), ##fail(B) : ##external(C,B), ##supp(A,B).
##wait(C,normal(B),I) :- ##internal(C,normal(B)), ##fail(normal(B)), ##steps(C,Z), I = 0..Z-1.
##wait(C,normal(B),I) :- ##internal(C,normal(B)), ##literal_tuple(B,A), ##wait(C,atom(A),I), ##steps(C,Z), I < Z.
##wait(C,sum(B,G),I)  :- ##internal(C,sum(B,G)), ##steps(C,Z), I = 0..Z-1, ##sum(B,G,T),
    #sum { W,L :   ##fail(atom(L)),   ##weighted_literal_tuple(B, L,W), L > 0, not ##scc(C,L) ;
           W,L : ##wait(C,atom(L),I), ##weighted_literal_tuple(B, L,W), L > 0,     ##scc(C,L) ;
           W,L :   ##true(atom(L)),   ##weighted_literal_tuple(B,-L,W), L > 0               } >= T-G+1.
##wait(C,atom(A),I)   :- ##wait(C,atom(A),0), ##steps(C,Z), I = 1..Z, ##wait(C,B,I-1) : ##supp(A,B), ##internal(C,B).

##bot :- ##scc(C,A), ##true(atom(A)), ##wait(C,atom(A),Z), ##steps(C,Z).

% saturate interpretations that are not answer sets

##true(atom(A)) :- ##supp(A), not ##fact(A), ##bot.
##fail(atom(A)) :- ##supp(A), not ##fact(A), ##bot.

%
% added by Javier
%

%#show.
%#show T : ##output(T,L), ##true(atom(L)).
#defined ##literal_tuple/1.
#defined ##literal_tuple/2.
#defined ##rule/2.
#defined ##atom_tuple/2.
#defined ##weighted_literal_tuple/3.
#defined ##scc/2.
"""

metaD_program_inc_base = """

% NOTE: assumes that a rule has no more than one head

##sum(B,G,T) :- ##rule(_,sum(B,G)), T = #sum { W,L : ##weighted_literal_tuple(B,L,W) }.

% extract supports of atoms and facts

##supp(A,B) :- ##rule(     choice(H),B), ##atom_tuple(H,A).
##supp(A,B) :- ##rule(disjunction(H),B), ##atom_tuple(H,A).

##supp(A) :- ##supp(A,_).

##atom(|L|) :- ##weighted_literal_tuple(_,L,_).
##atom(|L|) :- ##literal_tuple(_,L).
##atom( A ) :- ##atom_tuple(_,A).

##fact(A) :- ##rule(disjunction(H),normal(B)), ##atom_tuple(H,A), not ##literal_tuple(B,_).

% verify acyclic derivability

##internal(C,normal(B)) :- ##scc(C,A), ##supp(A,normal(B)), ##scc(C,A'), ##literal_tuple(B,A').
##internal(C,sum(B,G))  :- ##scc(C,A), ##supp(A,sum(B,G)),  ##scc(C,A'), ##weighted_literal_tuple(B,A',W).

##external(C,normal(B)) :- ##scc(C,A), ##supp(A,normal(B)), not ##internal(C,normal(B)).
##external(C,sum(B,G))  :- ##scc(C,A), ##supp(A,sum(B,G)),  not ##internal(C,sum(B,G)).

##steps(C,Z-1) :- ##scc(C,_), Z = { ##scc(C,A) : not ##fact(A) }.

%
% added by Javier
%

#defined ##literal_tuple/1.
#defined ##literal_tuple/2.
#defined ##rule/2.
#defined ##atom_tuple/2.
#defined ##weighted_literal_tuple/3.
#defined ##scc/2.
"""

metaD_program_parameters = ["m1", "m2"]

metaD_program_inc = """
% generate interpretation

##true(m1,m2,atom(A))                        :- ##fact(A),                not ##fixed(A). % added not fixed(A)
##true(m1,m2,atom(A)); ##fail(m1,m2,atom(A)) :- ##supp(A), not ##fact(A), not ##fixed(A). % added not fixed(A)
                       ##fail(m1,m2,atom(A)) :- ##atom(A), not ##supp(A), not ##fixed(A). % added not fixed(A)

##true(m1,m2,normal(B)) :- ##literal_tuple(B),
    ##true(m1,m2,atom(L)) : ##literal_tuple(B, L), L > 0;
    ##fail(m1,m2,atom(L)) : ##literal_tuple(B,-L), L > 0.
##fail(m1,m2,normal(B)) :- ##literal_tuple(B, L), ##fail(m1,m2,atom(L)), L > 0.
##fail(m1,m2,normal(B)) :- ##literal_tuple(B,-L), ##true(m1,m2,atom(L)), L > 0.

##true(m1,m2,sum(B,G)) :- ##sum(B,G,T),
    #sum { W,L : ##true(m1,m2,atom(L)), ##weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : ##fail(m1,m2,atom(L)), ##weighted_literal_tuple(B,-L,W), L > 0 } >= G.
##fail(m1,m2,sum(B,G)) :- ##sum(B,G,T),
    #sum { W,L : ##fail(m1,m2,atom(L)), ##weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : ##true(m1,m2,atom(L)), ##weighted_literal_tuple(B,-L,W), L > 0 } >= T-G+1.

% verify supported model properties

##bot(m1,m2) :- ##rule(disjunction(H),B), ##true(m1,m2,B), ##fail(m1,m2,atom(A)) : ##atom_tuple(H,A).
##bot(m1,m2) :- ##true(m1,m2,atom(A)), ##fail(m1,m2,B) : ##supp(A,B).

% verify acyclic derivability

##wait(m1,m2,C,atom(A),0)   :- ##scc(C,A), ##fail(m1,m2,B) : ##external(C,B), ##supp(A,B).
##wait(m1,m2,C,normal(B),I) :- ##internal(C,normal(B)), ##fail(m1,m2,normal(B)), ##steps(C,Z), I = 0..Z-1.
##wait(m1,m2,C,normal(B),I) :- ##internal(C,normal(B)), ##literal_tuple(B,A), ##wait(m1,m2,C,atom(A),I), ##steps(C,Z), I < Z.
##wait(m1,m2,C,sum(B,G),I)  :- ##internal(C,sum(B,G)), ##steps(C,Z), I = 0..Z-1, ##sum(B,G,T),
    #sum { W,L :   ##fail(m1,m2,atom(L)),   ##weighted_literal_tuple(B, L,W), L > 0, not ##scc(C,L) ;
           W,L : ##wait(m1,m2,C,atom(L),I), ##weighted_literal_tuple(B, L,W), L > 0,     ##scc(C,L) ;
           W,L :   ##true(m1,m2,atom(L)),   ##weighted_literal_tuple(B,-L,W), L > 0               } >= T-G+1.
##wait(m1,m2,C,atom(A),I)   :- ##wait(m1,m2,C,atom(A),0), ##steps(C,Z), I = 1..Z, ##wait(m1,m2,C,B,I-1) : ##supp(A,B), ##internal(C,B).

##bot(m1,m2) :- ##scc(C,A), ##true(m1,m2,atom(A)), ##wait(m1,m2,C,atom(A),Z), ##steps(C,Z).

% saturate interpretations that are not answer sets

##true(m1,m2,atom(A)) :- ##supp(A), not ##fact(A), ##bot(m1,m2), not ##fixed(A).
##fail(m1,m2,atom(A)) :- ##supp(A), not ##fact(A), ##bot(m1,m2), not ##fixed(A).

%
% added by Javier
%

#defined ##literal_tuple/1.
#defined ##literal_tuple/2.
#defined ##rule/2.
#defined ##atom_tuple/2.
#defined ##weighted_literal_tuple/3.
#defined ##scc/2.

"""





# MIT License
#
# Copyright (c) 2017 Javier Romero
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-


# imports
import subprocess
import tempfile
import re


# defines
CLINGO = "clingo"
REIFY_OUTPUT = "--output=reify"
SMODELS_OUTPUT = "--output=smodels"
LP2NORMAL = "lp2normal2"
LP2SAT = "lp2sat"
VERSION = "--version"
NO_CLINGO = "clingo binary version not found (when running 'clingo --version')"
OLD_CLINGO = """clingo binary too old (when running 'clingo --version')\
 version 5.3 or newer is needed"""

# options for translations to SAT
LP2NORMAL_OPTIONS = []
LP2SAT_OPTIONS = []

#
# classes Node and Graph used by reify_from_observer()
#

# by Roland Kaminski, modified by Javier Romero
class Node:

    def __init__(self, name):
        self.name = name
        self.neighbors = []
        self.visited_  = 0
        self.finished_ = 1

    def add_neighbor(self, neighbor): self.neighbors.append(neighbor)
    def get_neighbors(self): return self.neighbors
    def next_neighbors(self): return self.neighbors[self.finished_ - 1:]

    def visit(self, visited): self.visited_ = visited + 1
    def visited(self): return self.visited_ - 1

    def mark(self): self.visited_ = 1
    def marked(self): return self.visited_ >= 1

    def pop(self): self.finished_ = 0
    def popped(self): return self.finished_ == 0
    def inc_finished(self): self.finished_ += 1


# by Roland Kaminski, modified by Javier Romero
class Graph:

    def __init__(self):
        self.graph = {}
        self.sccs = []
        self.singletons = set()

    def add_edge(self, head, body):
        if head == body:
            self.singletons.add(head)
            return
        node = self.graph.get(head, None)
        if not node:
            node = Node(head)
            self.graph[head] = node
        node.add_neighbor(body)

    def reify_sccs(self, prefix=""):
        # compute sccs and update self.singletons
        for _, v in self.graph.items():
            if not v.marked():
                self.tarjan(v)
        # return string
        out = ""
        for idx, item in enumerate(list(self.singletons) + self.sccs):
            out += " ".join(["{}scc({},{}).".format(prefix, idx, i) for i in item]) + "\n"
        return out

    # private
    def next(self, v):
        for y in v.next_neighbors():
            v.inc_finished()
            node = self.graph.get(y, None)
            if node and not node.marked():
                return node
        return None

    # private
    def root(self, v):
        root = True
        for y in v.get_neighbors():
            node = self.graph.get(y, None)
            if node and not node.popped() and node.visited() < v.visited():
                root = False
                v.visit(node.visited())
        return root

    # private
    def tarjan(self, start):

        s, t = [], []
        visited = 0
        s.append(start)
        start.mark()

        def top(s): return s[-1]

        while len(s) > 0:
            x = top(s)
            if x.visited() == 0:
                visited  += 1
                x.visit(visited)
            y = self.next(x)
            if y != None:
                s.append(y)
                y.mark()
            else:
                s.pop()
                if self.root(x):
                    scc = [x.name]
                    x.pop()
                    while len(t) > 0 and top(t).visited() >= x.visited():
                        y = t.pop()
                        y.pop()
                        scc.append(y.name)
                        self.singletons.discard(y.name) # update self.singletons
                    if len(scc)>1:
                        self.sccs.append(scc)           # update self.sccs
                        self.singletons.discard(x.name) # update self.singletons
                else: t.append(x)


#
# reify_from_observer()
#
# * uses a clingo observer with fields:
#   - rules
#   - weight_rules
#   - output_atoms
#   - output_terms
#

# TODO: Implement option where we take care about repeated heads and bodies
def reify_from_observer(observer, prefix=""):

    # start
    out = ""
    literal_tuple, wliteral_tuple, atom_tuple = 1, 1, 1
    p = prefix

    # fact 0
    # out += "% fact 0\n"
    out += "{}rule(disjunction(0),normal(0)). ".format(p)
    out += "{}atom_tuple(0,0). {}literal_tuple(0).\n\n".format(p, p)

    # start graph
    graph = Graph()

    # normal rules
    for choice, head, body in observer.rules:

        # out += "% normal rule\n"
        # body
        if body:
            out += "{}literal_tuple({}).".format(p, literal_tuple) + "\n"
            out += " ".join([
                "{}literal_tuple({},{}).".format(p, literal_tuple, l)
                for l in body
            ]) + "\n"
        # head
        head_type = "choice" if choice else "disjunction"
        out += "{}rule({}({}),normal({})).".format(
            p, head_type, atom_tuple, literal_tuple if body else 0
        ) + "\n"
        out += " ".join([
            "{}atom_tuple({},{}).".format(p, atom_tuple, l) for l in head
        ]) + "\n\n"
        # update counters
        if body:
            literal_tuple += 1
        atom_tuple += 1
        # update graph (this can be interwined with the rules above)
        for l in body:
            if l >= 0:
                for atom in head:
                    graph.add_edge(atom, l)

    # weight rules
    for choice, head, lower_bound, body in observer.weight_rules:
        # out += "% weighted rule\n"
        # body
        out += " ".join([
            "{}weighted_literal_tuple({},{},{}).".format(
                p, wliteral_tuple, l, w
            ) for l, w in body
        ]) + "\n"
        # head
        head_type = "choice" if choice else "disjunction"
        out += "{}rule({}({}),sum({},{})).".format(
            p, head_type, atom_tuple, wliteral_tuple, lower_bound
        ) + "\n"
        out += " ".join([
            "{}atom_tuple({},{}).".format(p, atom_tuple, l) for l in head
        ]) + "\n\n"
        # update counters
        wliteral_tuple += 1
        atom_tuple += 1
        # update graph (this can be interwined with the rules above)
        for l, w in body:
            if l >= 0:
                for atom in head:
                    graph.add_edge(atom, l)

    # sccs
    # out += "% sccs\n"
    out += graph.reify_sccs(p) + "\n"

    # output atoms
    # out += "% output atoms\n"
    for symbol, atom in observer.output_atoms:
        out += "{}output({},{}).\n".format(p, symbol, atom)

    # output terms
    # out += "% output term\n"
    for symbol, condition in observer.output_terms:
        out += "{}output_term({}).\n".format(p, symbol) + "\n"
        out += " ".join([
            "{}output_term({},{}).".format(p, symbol, l) for l in condition
        ]) + "\n"

    return out


#
# reify_from_string()
#
# * uses a clingo binary not older than than clingo-5.3
#

# run command and return stdout as a string
def run_command(command):
    output = subprocess.check_output(command)
    if isinstance(output, bytes):
        return output.decode()
    return output

def check_clingo_version():
    version = run_command([CLINGO, VERSION])
    match = re.match(r'clingo version (\d+)\.(\d+)', version)
    if match:
        first  = int(match.group(1))
        second = int(match.group(2))
        if first > 5 or (first == 5 and second >= 3):
            return
        raise Exception(OLD_CLINGO)
    raise Exception(NO_CLINGO)

def reify_from_string(program, prefix):
    check_clingo_version()
    with tempfile.NamedTemporaryFile(delete=False) as file_in:
        # write program to file_in
        file_in.write(program.encode())
        file_in.flush()
    # run command
    command = [CLINGO, REIFY_OUTPUT, file_in.name]
    output = run_command(command)
    # add prefix and return
    output = re.sub(r'^(\w+)', r'' + prefix + r'\1', output, flags=re.M)
    return output


#
# reify_from_string_through_sat()
#
# * uses a clingo binary, lp2normal2 and lp2sat
#

CHOICE_FACTS = """\
{0}rule(choice(0),normal(0)). {0}literal_tuple(0). {0}atom_tuple(0,1..{1}).
"""

OUTPUT_FACTS = """\
{0}output({1},{2}). {0}literal_tuple({2}). {0}literal_tuple({2},{3}).
"""

def reify_from_string_through_sat(program, prefix):

    # translate to dimacs
    check_clingo_version()
    # write program to file_in
    with tempfile.NamedTemporaryFile(delete=False) as file_in:
        file_in.write(program.encode())
        file_in.flush()
    # run commands
    ps1 = subprocess.Popen(
        [CLINGO, SMODELS_OUTPUT, file_in.name], stdout=subprocess.PIPE
    )
    ps2 = subprocess.Popen(
        [LP2NORMAL] + LP2NORMAL_OPTIONS,
        stdin=ps1.stdout, stdout=subprocess.PIPE
    )
    dimacs = subprocess.check_output(
        [LP2SAT] + LP2SAT_OPTIONS, stdin=ps2.stdout
    )
    if isinstance(dimacs, bytes):
        dimacs = dimacs.decode()

    # generate output
    output = ""
    literal_tuples = 1
    for line in dimacs.splitlines():

        # first line
        match = re.match(r'p\s+cnf\s+(\d+)\s(\d+)\s*\Z', line)
        if match:
            atoms = int(match.group(1))
            if atoms > 0:
                output += CHOICE_FACTS.format(prefix, atoms)
            continue

        # comments
        match = re.match(r'c\s+(\d+)\s(.*)\Z', line)
        if match:
            output += OUTPUT_FACTS.format(
                prefix, match.group(2), literal_tuples, match.group(1)
            )
            literal_tuples += 1
            continue
        if line[0] == 'c':
            continue

        # clauses
        output += "{0}rule(disjunction(1),normal({1})).".format(
            prefix, literal_tuples
        )
        output += " {0}literal_tuple({1}).".format(prefix, literal_tuples)
        for lit in line.split():
            if lit != "0":
                output += " {0}literal_tuple({1},{2}).".format(
                    prefix, literal_tuples,
                    lit[1:] if lit[0] == "-" else "-" + lit
                )
        output += "\n"
        literal_tuples += 1

    # return output
    return output




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

import os
import tempfile
import re
import sys
import clingo

#
# defines
#

# programs
EMPTY      = ""
BASE       = "base"
SPEC       = "specification"
GENERATE   = "generate"
PREFP      = "preference"
PBASE      = "preference_base"
HEURISTIC  = "heuristic"
APPROX     = "weak"
UNSATP     = "preference_unsat"
UNSATPBASE = "preference_unsat_base"
CONSTANTS_NB = "constants_nonbase"
ON_OPT_HEUR_PROGRAM  = "on_opt_heuristic"
METAPROGRAM = "metaprogram"
METAUNSAT = "meta_unsat"
METAUNSAT_BASE = "meta_unsat_base"
# query program name and program
QUERY         = "query"
QUERY_PROGRAM = ":- not query."

# map preference programs to their bases
mapbase = { PREFP  : PBASE,
            UNSATP : UNSATPBASE }

# underscores for programs
U_PREFP     = 1
U_APPROX    = 2
U_HEURISTIC = 3
U_UNSATP    = 4
U_METABASE  = 5
U_METAPREF  = 6

# predicate names
DOM        = "dom" 
GEN_DOM    = "gen_dom"
VOLATILE   = "volatile"
MODEL      = "m"
HOLDS      = "holds" 
HOLDSP     = "holds'"
PREFERENCE = "preference"
OPTIMIZE   = "optimize"
UNSAT      = "unsat"
ERROR_PRED = "error"
WARN_PRED  = "warning"
SHOW       = "show"
LAST_HOLDS = "last_holds"
LAST_SHOWN = "last_shown"
HOLDS_AT_ZERO = "holds_at_zero"

# translation tokens
HASH_SEM = "#sem"

# ast
NO_SIGN = clingo.ast.Sign.NoSign

# errors
ERROR_PROJECT = """\
error: syntax error, unexpected #project statement in {} program
  {}\n"""
ERROR_MINIMIZE = """\
error: syntax error, unexpected clingo optimization statement in {} program
  {}\n"""
ERROR_HEURISTIC = """\
error: syntax error, unexpected clingo heuristic directive in {} program
  {}\n"""
ERROR_DISJOINT = """\
error: syntax error, unexpected disjoint atom in {} program
  {}\n"""
ERROR_CSPLITERAL = """\
error: syntax error, unexpected csp literal in {} program
  {}\n"""
ERROR_KEYWORD = """\
error: syntax error, special predicate depends on non domain atoms in {} program
  {}/{}\n"""
ERROR_HOLDSP = """\
error: syntax error, unexpected holds'/1 predicate in {} program
  {}\n"""
ERROR_HOLDSP_H = """\
error: syntax error, unexpected holds'/1 predicate in the head of a heuristic \
directive in a {} program
  {}\n"""
ERROR_HOLDS_H = """\
error: syntax error, holds/1 predicate in the head of a heuristic \
directive in a {} program depends on non domain atoms
  {}\n"""

# used by option --on-opt-heur
POS        = 0
NEG        = 1
SHOWN_ATOM = 0
PREF_ATOM  = 1

# used by option --meta
META_OPEN    = 0
META_NO      = 1
META_SIMPLE  = 2
META_COMBINE = 3

# messages (used by solver and controller)
STR_MODEL_FOUND        = "MODEL FOUND"
STR_MODEL_FOUND_STAR   = "MODEL FOUND *"
SATISFIABLE            = "SATISFIABLE"

#
# global variables
#
underscores = ""

#
# classes
#

# check program_parser for an usage example
class Capturer:
    
    def __init__(self,stdx):
        self.__original_fd      = stdx.fileno()
        self.__save_original_fd = os.dup(self.__original_fd)
        self.__tmp_file         = tempfile.TemporaryFile()
        os.dup2(self.__tmp_file.fileno(), self.__original_fd)

    def read(self):
        self.__tmp_file.flush()
        self.__tmp_file.seek(0)
        out = self.__tmp_file.read()
        if isinstance(out, bytes):
            out = out.decode()
        return out

    def close(self):
        os.dup2(self.__save_original_fd, self.__original_fd)
        os.close(self.__save_original_fd)
        self.__tmp_file.close()

    def translate_error(self, programs, program, string):
        if program != BASE:
            return string
        out = ""
        for i in string.splitlines():
            printed = False
            match = re.match(r'<block>:(\d+):(\d+)-(\d+:)?(\d+): (.*)',i)
            if match:
                # get parsing
                error_line = int(match.group(1))
                col_ini    = int(match.group(2))
                if match.group(3) is not None:
                    line_extra = int(match.group(3)[:-1]) 
                else:
                    line_extra = None
                col_end    = int(match.group(4))
                rest       =     match.group(5)
                # for every position
                positions = programs[program][""].get_positions()
                for pos in positions:
                    if pos.lines >= error_line:
                        # set Location attributes
                        if error_line == 1:
                            col_ini += pos.col - 1
                            if not line_extra:
                                col_end += pos.col - 1
                        loc_line = error_line + pos.line - 1
                        if line_extra is None:
                            loc_line_extra = loc_line
                        else:
                            loc_line_extra = line_extra + pos.line - 1
                        # create Location and print
                        loc = Location(pos.filename, loc_line, col_ini, 
                                       loc_line_extra, col_end)
                        out += "{}{}\n".format(loc,rest)
                        printed = True
                        break
                    else:
                        error_line = error_line - pos.lines
                        if line_extra:
                            line_extra = line_extra - pos.lines
                        else:
                            line_extra = None
            if not printed:
                out += i + "\n"
        # for
        return out

class SilentException(Exception):
    pass

class FatalException(Exception):
    pass

#
# Location, and ProgramPosition
#

class Location(object):
    
    def __init__(self, filename, line, col_ini, line_extra, col_end):
        self.filename   = filename
        self.line       = line
        self.col_ini    = col_ini
        self.line_extra = line_extra
        self.col_end    = col_end

    def get_position(self):
        return ProgramPosition(self.filename, self.line,
                               self.col_ini, self.line_extra - self.line)

    def __repr__(self):
        if not self.line_extra or self.line_extra == self.line:
            extra = ""
        else:
            extra = "{}:".format(self.line_extra)
        args = (self.filename, self.line, self.col_ini, extra, self.col_end)
        return "{}:{}:{}-{}{}: ".format(*args)

# TODO: get rid of this, and use just Location
class ProgramPosition(object):

    def __init__(self, filename, line, col, lines=1):
        self.filename = filename
        self.line     = line
        self.col      = col
        self.lines    = lines # number of lines




# MIT License
#
# Copyright (c) 2017 Javier Romero
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-


#
# IMPORTS
#

import clingo
import re
from . import metasp_programs
from . import reify
from ...utils import utils


#
# DEFINES
#

U_METABASE = utils.U_METABASE
U_METAPREF = utils.U_METAPREF
PREFERENCE = utils.PREFERENCE
OPTIMIZE   = utils.OPTIMIZE
VOLATILE      = utils.VOLATILE
HOLDS         = utils.HOLDS
UNSAT_ATOM    = utils.UNSAT
HOLDS_AT_ZERO = utils.HOLDS_AT_ZERO

METAPREF_BASIC = """
{ ##""" + HOLDS + """(X,0..1) } :- X = @get_holds_domain().
##""" + VOLATILE + """(##m(0),##m(1)).
:- ##""" + UNSAT_ATOM + """(##m(0),##m(1)).
#show ##""" + HOLDS + """/2.
#const ##m1=0.
#const ##m2=1.
"""

BINDING_SIMPLE_PYTHON = """
**true(atom(A)) :-     ##""" + HOLDS + """(X,0), **output(##""" + HOLDS + """(X,1),A).                    % from base
**fail(atom(A)) :- not ##""" + HOLDS + """(X,0), **output(##""" + HOLDS + """(X,1),A).                    % from base
**true(atom(A)) :- $$output_term(##""" + HOLDS_AT_ZERO + """(X)), **output(##""" + HOLDS + """(X,0),A),   % from meta_base
                   $$true(atom( B)) : $$output_term(##""" + HOLDS_AT_ZERO + """(X),B), B >= 0;
                   $$fail(atom(-B)) : $$output_term(##""" + HOLDS_AT_ZERO + """(X),B), B <  0.
**fail(atom(A)) :- $$output_term(##""" + HOLDS_AT_ZERO + """(X)), **output(##""" + HOLDS + """(X,0),A),   % from meta_base
                   $$fail(atom( B)),  $$output_term(##""" + HOLDS_AT_ZERO + """(X),B), B >= 0.
**fail(atom(A)) :- $$output_term(##""" + HOLDS_AT_ZERO + """(X)), **output(##""" + HOLDS + """(X,0),A),   % from meta_base
                   $$true(atom(-B)),  $$output_term(##""" + HOLDS_AT_ZERO + """(X),B), B <  0.
**bot :- $$bot.
$$bot :- **bot.
:- not **bot.
$$atom(|B|) :- $$output_term(X,B). % needed when X is a fact
"""

BINDING_SIMPLE_BINARY = """
**true(atom(A)) :-     ##""" + HOLDS + """(X,0), **output(##""" + HOLDS + """(X,1),B), **literal_tuple(B,A).  % from base
**fail(atom(A)) :- not ##""" + HOLDS + """(X,0), **output(##""" + HOLDS + """(X,1),B), **literal_tuple(B,A).  % from base
**true(atom(A)) :- $$true(normal(B)), $$output(##""" + HOLDS + """(X,0),B),                       % from meta_base
                   **output(##""" + HOLDS + """(X,0),C), **literal_tuple(C,A).
**fail(atom(A)) :- $$fail(normal(B)), $$output(##""" + HOLDS + """(X,0),B),                       % from meta_base
                   **output(##""" + HOLDS + """(X,0),C), **literal_tuple(C,A).
**bot :- $$bot.
$$bot :- **bot.
:- not **bot.
"""

BINDING_INC_PYTHON_BASE = """
##fixed(A) :- ##output($$""" + HOLDS + """(X,1),A).
"""

BINDING_INC_BINARY_BASE = """
##fixed(A) :- ##output($$""" + HOLDS + """(X,1),B), ##literal_tuple(B,A).
"""

BINDING_INC_PYTHON = """
##true(m1,m2,atom(A)) :-     $$""" + HOLDS + """(X,m2), ##output($$""" + HOLDS + """(X,1),A).
##fail(m1,m2,atom(A)) :- not $$""" + HOLDS + """(X,m2), ##output($$""" + HOLDS + """(X,1),A).
##true(m1,m2,atom(A)) :-     $$""" + HOLDS + """(X,m1), ##output($$""" + HOLDS + """(X,0),A).
##fail(m1,m2,atom(A)) :- not $$""" + HOLDS + """(X,m1), ##output($$""" + HOLDS + """(X,0),A).
$$""" + UNSAT_ATOM + """($$m(m1),$$m(m2)) :- $$""" + VOLATILE + """($$m(m1),$$m(m2)).
:- not ##bot(m1,m2).
"""

BINDING_INC_BINARY = """
##true(m1,m2,atom(A)) :-     $$""" + HOLDS + """(X,m2), ##output($$""" + HOLDS + """(X,1),B), ##literal_tuple(B,A).
##fail(m1,m2,atom(A)) :- not $$""" + HOLDS + """(X,m2), ##output($$""" + HOLDS + """(X,1),B), ##literal_tuple(B,A).
##true(m1,m2,atom(A)) :-     $$""" + HOLDS + """(X,m1), ##output($$""" + HOLDS + """(X,0),B), ##literal_tuple(B,A).
##fail(m1,m2,atom(A)) :- not $$""" + HOLDS + """(X,m1), ##output($$""" + HOLDS + """(X,0),B), ##literal_tuple(B,A).
$$""" + UNSAT_ATOM + """($$m(m1),$$m(m2)) :- $$""" + VOLATILE + """($$m(m1),$$m(m2)).
:- not ##bot(m1,m2).
"""

BINDING_INC_VOLATILE = """
##true(m1,m2,atom(A)) :- ##supp(A), not ##fact(A), not ##fixed(A), not $$""" + VOLATILE + """($$m(m1),$$m(m2)).
##fail(m1,m2,atom(A)) :- ##supp(A), not ##fact(A), not ##fixed(A), not $$""" + VOLATILE + """($$m(m1),$$m(m2)).
"""

# get_holds_domain() should not be defined here
ASPRIN_LIBRARY_PY = """
#script(python)

import math

def exp2(x):
    return int(math.pow(2,x.number))

def get(atuple, index):
    try:
        return atuple.arguments[index.number]
    except:
        return atuple

def get_mode():
    return 'normal'

sequences = {}
def get_sequence(name, elem):
    string = str(name)
    if string in sequences:
        sequences[string] += 1
    else:
        sequences[string]  = 1
    return sequences[string]

def length(atuple):
    try:
        return len(atuple.arguments)
    except:
        return 1 

def log2up(x):
    return int(math.ceil(math.log(x.number,2)))

#end.
"""

HOLDS_DOMAIN_PY = """
import clingo
holds_domain = {}
def get_holds_domain():
    return holds_domain
"""

class Observer:

    def __init__(
        self,
        control,
        register_observer = False,
        replace = False,
        bool_add_statement = False,
        bool_add_base = False,
        bool_add_specification = False,
        bool_add_constants_nb = False
    ):
        # flags
        self.bool_add_statement     = bool_add_statement
        self.bool_add_base          = bool_add_base
        self.bool_add_specification = bool_add_specification
        self.bool_add_constants_nb  = bool_add_constants_nb
        if register_observer:
            control.register_observer(self, replace)
        # observations
        self.rules         = []
        self.weight_rules  = []
        self.output_atoms  = []
        self.output_terms  = []
        self.statements    = []
        self.base          = None         # (program, old consts, new consts)
        self.specification = None         # (program, old consts, new consts)
        self.constants_nb  = ("", [], []) # (program,         [],         [])

    #
    # control object observer
    #

    def rule(self, choice, head, body):
        self.rules.append((choice, head, body))

    def weight_rule(self, choice, head, lower_bound, body):
        self.weight_rules.append((choice, head, lower_bound, body))

    def output_atom(self, symbol, atom):
        self.output_atoms.append((symbol, atom))

    def output_term(self, symbol, condition):
        self.output_terms.append((symbol, condition))

    #
    # program parser observer
    #

    def add_statement(self, statement):
        if self.bool_add_statement:
            self.statements.append(statement)

    def add_base(self, program, old, new):
        if self.bool_add_base:
            self.base = (program, old, new)

    def add_specification(self, program, old, new):
        if self.bool_add_specification:
            self.specification = (program, old, new)

    def add_constants_nb(self, program, old, new):
        if self.bool_add_constants_nb:
            self.constants_nb = (program, old, new)


# abstract class
class AbstractMetasp:

    def __init__(self, solver):
        # uses solver.control, solver.observer and solver.underscores
        self.solver = solver
        self.binding_simple = None   # to be defined by subclasses
        self.binding_inc_base = None # to be defined by subclasses
        self.binding_inc = None      # to be defined by subclasses

    # private
    # to be defined by subclasses (used by get_incremental_program())
    def get_meta_pref_facts(self, prefix):
        return None

    # private
    # to be defined by subclasses (used by get_meta_program())
    def get_meta_base_facts(self, prefix):
        return None

    # public
    # uses get_meta_base_facts(), get_meta_pref_facts and binding0
    def get_meta_program(self):
        # base
        prefix = self.solver.underscores + "_"*U_METABASE
        meta_base  = self.get_meta_base_facts(prefix)
        meta_base += metasp_programs.metaD_program.replace("##", prefix)
        # pref
        prefix = self.solver.underscores + "_"*U_METAPREF
        meta_pref = self.get_meta_pref_facts(prefix)
        meta_pref += metasp_programs.metaD_program.replace("##", prefix)
        # binding
        meta_bind = self.get_meta_bind(self.binding_simple)
        # return
        return meta_base + meta_pref + meta_bind

    # public
    # uses get_meta_pref_facts, binding_inc and binding_inc_base
    def get_incremental_program(self):
        u = self.solver.underscores
        prefix = u + "_"*U_METAPREF
        # base
        base = self.get_meta_pref_facts(prefix)
        base += metasp_programs.metaD_program_inc_base.replace("##", prefix)
        base += self.binding_inc_base.replace("$$", u).replace("##", prefix)
        # parameters
        parameters = metasp_programs.metaD_program_parameters
        # incremental
        inc = metasp_programs.metaD_program_inc.replace("##", prefix)
        binding = self.binding_inc + BINDING_INC_VOLATILE
        inc += binding.replace("$$", u).replace("##", prefix)
        # return
        return base, parameters, inc

    #
    # get_pref() (used by MetaspPython and MetaspBinary)
    #

    def statement_to_str(self, statement):
        if str(statement.type) == "Definition": # to avoid printing [default]
            return "#const {}={}.".format(statement.name, statement.value)
        elif str(statement.type) == "Program":
            return "" # IMPORTANT: program statements are skipped
        return str(statement)

    def get_specification(self):
        underscores = self.solver.underscores
        signatures = [(underscores + PREFERENCE, 2),
                      (underscores + PREFERENCE, 5),
                      (underscores +   OPTIMIZE, 1)]
        symbolic_atoms = self.solver.control.symbolic_atoms
        spec = ""
        for name, arity in signatures:
            spec += " ".join([
                str(atom.symbol) + "."
                for atom in symbolic_atoms.by_signature(name, arity)
            ]) + "\n"
        return spec

    # WARNING:
    # This is incorrect in general, since it may change the
    # content of some strings.
    # This is needed so that @functions of 0 parameters are printed
    # correctly. This should be fixed in clingo.
    def fix_functions(self, program):
        pattern = r'(@_*[a-z][\'A-Za-z0-9_]*)([^\(\'A-Za-z0-9_])'
        return re.sub(pattern, r'\1()\2', program)

    def get_pref(self):
        # basic program
        basic = METAPREF_BASIC.replace("##", self.solver.underscores)
        # preference specification
        specification = self.get_specification()
        # preference program
        preference_program = "\n".join([
            self.statement_to_str(s) for s in self.solver.observer.statements
        ])
        # See WARNING above
        preference_program = self.fix_functions(preference_program)
        # constants
        constants = self.solver.observer.constants_nb[0]
        # return
        return basic + specification + preference_program + constants

    #
    # get_meta_bind() (used by get_meta_program())
    #

    def get_meta_bind(self, binding):
        out = binding.replace("##", self.solver.underscores)
        prefix_base = self.solver.underscores + "_"*U_METABASE
        prefix_pref = self.solver.underscores + "_"*U_METAPREF
        return out.replace("$$", prefix_base).replace("**", prefix_pref)


# Uses the observer in Python
class MetaspPython(AbstractMetasp):

    def __init__(self, solver):
        AbstractMetasp.__init__(self, solver)
        self.binding_simple = BINDING_SIMPLE_PYTHON
        self.binding_inc_base = BINDING_INC_PYTHON_BASE
        self.binding_inc = BINDING_INC_PYTHON

    def get_meta_base_facts(self, prefix):
        return reify.reify_from_observer(self.solver.observer, prefix)

    def get_meta_pref_facts(self, prefix):
        ctl = clingo.Control([])
        observer = Observer(ctl, register_observer=True, replace=True)
        ctl.add("base", [], self.get_pref())
        ctl.ground([("base",[])], self.solver)
        return reify.reify_from_observer(observer, prefix)


# Uses clingo binary
class MetaspBinary(AbstractMetasp):

    def __init__(self, solver):
        AbstractMetasp.__init__(self, solver)
        self.binding_simple = BINDING_SIMPLE_BINARY
        self.binding_inc_base = BINDING_INC_BINARY_BASE
        self.binding_inc = BINDING_INC_BINARY

    def call_reify(self, program, prefix):
        return reify.reify_from_string(program, prefix)

    # WARNING:
    # The next function relies on the form of the specification programs,
    # assumes rules for preference/2, preference/5, optimize/1
    # start at the beginning of a line and finish at the end.
    # Depends on spec_parser/ast.py.
    def get_meta_base_facts(self, prefix):
        observer = self.solver.observer
        # get base
        program  = observer.base[0] + "\n"
        # get specification without preference/2, /5 and optimize/1
        pattern = r'^(\_*)(' + PREFERENCE + r'|' + OPTIMIZE + r')'
        program += re.sub(
            pattern, r'%\1\2', observer.specification[0], flags=re.M
        ) + "\n"
        # add constants
        for idx, old in enumerate(observer.base[1]): # for every old constant
            program += "#const {}={}.".format(old, observer.base[2][idx]) + "\n"
        # add show
        program += "#show " + self.solver.underscores + "holds/2.\n"
        # get meta using clingo binary
        return reify.reify_from_string(program, prefix)

    def get_meta_pref_facts(self, prefix):
        holds_domain = ",\n".join([
            '  clingo.parse_term("""{}""")'.format(x) 
            for x in self.solver.holds_domain
        ])
        get_holds_domain = HOLDS_DOMAIN_PY.format("[\n" + holds_domain + "\n]")
        library = ASPRIN_LIBRARY_PY.replace(
            "#end.", get_holds_domain + "\n#end."
        )
        prefix = self.solver.underscores + "_"*U_METAPREF
        return self.call_reify(self.get_pref() + library, prefix)

# Uses clingo binary, lp2normal2 and lp2sat
class MetaspSAT(MetaspBinary):

    def __init__(self, solver):
        MetaspBinary.__init__(self, solver)

    def call_reify(self, program, prefix):
        return reify.reify_from_string_through_sat(program, prefix)

